import pysam
from sinto import utils
from scipy import sparse
import numpy as np
from collections import Counter, defaultdict
from multiprocessing import Pool
import functools
import re
import gc
import tempfile
import os
import math


def writeFragments(fragments, filepath):
    """Write fragments to file

    Parameters
    ----------
    fragments : list
        3-deep list of ATAC fragments
    filepath : str
        Path for output file
    """
    with open(filepath, "a") as outf:
        for i in fragments:
            outstr = "\t".join(map(str, i))
            outf.write(outstr + "\n")


def createPositionLookup(frags, pos=1):
    """Create dictionary where key is subsetted fragment coords
    (no start or no stop), value is the list of full fragments
    that share the coordinates in the key.
    Only entries where >1 full fragments share the same coordinate
    are retained"""
    fragsplit = [x.split("|") for x in frags]
    posfrags = ["|".join([x[0], x[pos], x[3]]) for x in fragsplit]
    counts = Counter(posfrags)
    starts = dict()
    for i in range(len(frags)):
        if counts[posfrags[i]] > 1:
            try:
                starts[posfrags[i]]
            except KeyError:
                starts[posfrags[i]] = [frags[i]]
            else:
                starts[posfrags[i]].append(frags[i])
        else:
            pass
    return starts


def collapseOverlapFragments(counts, pos=1):
    """Collapse fragment counts that share a common start
    or end coordinate and are from the same cell"""
    keys = list(counts.keys())
    startfrags = createPositionLookup(frags=keys, pos=pos)
    for i in startfrags.keys():
        # extract counts for all fragments sharing start coord
        fullfrags = startfrags[i]
        countvec = [counts[x] for x in fullfrags]
        # find which fragment has the max
        # if ties, choose first in tie
        windex = countvec.index(max(countvec))
        winner = fullfrags[windex]
        # update count for the winner
        counts[winner] = sum(countvec)
        # remove losers
        del fullfrags[windex]
        for j in fullfrags:
            del counts[j]
    return counts


def collapseFragments(fragments, collapse_within):
    """Collapse duplicate fragments
    """
    fraglist = list(fragments.values())
    fragcoords_with_bc = ["|".join(map(str, x)) for x in fraglist]
    counts = Counter(fragcoords_with_bc)

    if len(fraglist) == 0:
        return list()

    # enumerate fragments and barcodes
    frag_id_lookup = id_lookup(l=["|".join(map(str, x[:3])) for x in fraglist])
    bc_id_lookup = id_lookup(l=[x[3] for x in fraglist])

    # collapse counts from the same cell barcode with partial overlap
    counts = collapseOverlapFragments(counts, pos=1)
    counts = collapseOverlapFragments(counts, pos=2)
    
    # if collapsing within barcodes, we skip the following steps
    if collapse_within:
        counts = list(counts.items())
        cv = [x[0].split("|") for x in counts]
        for i in range(0, len(counts)):
            cv[i].append(counts[i][1])
        return cv

    # get list of barcode index and fragment index from counts
    row = []
    col = []
    data = []
    for i in list(counts.items()):
        data.append(i[1])
        rowstr = i[0].split("|")[:3]
        bcstr = i[0].split("|")[3]
        row.append(frag_id_lookup["|".join(rowstr)])
        col.append(bc_id_lookup[bcstr])

    # free memory
    del fraglist
    del fragments
    del counts
    gc.collect()

    # create sparse matrix of fragment counts from fraglist (column, row, value)
    mat = sparse.coo_matrix(
        (data, (np.array(row), np.array(col))),
        shape=(max(frag_id_lookup.values()) + 1, max(bc_id_lookup.values()) + 1),
    )
    mat = mat.tocsr()

    # find which barcode contains the most counts for each fragment
    rowmax = np.argmax(mat, axis=1)
    rowsum = mat.sum(axis=1).tolist()

    # collapse back into a list of fragment coords and barcodes
    frag_inverse = dict(zip(frag_id_lookup.values(), frag_id_lookup.keys()))
    bc_inverse = dict(zip(bc_id_lookup.values(), bc_id_lookup.keys()))
    collapsed_frags = [frag_inverse[x] for x in np.arange(mat.shape[0])]
    collapsed_barcodes = [bc_inverse[x[0]] for x in rowmax.tolist()]
    collapsed = []
    for i in range(len(collapsed_barcodes)):
        if rowsum[i][0] < 1:
            continue
        else:
            frag = collapsed_frags[i].split("|")
            frag.append(collapsed_barcodes[i])
            frag.append(rowsum[i][0])
            collapsed.append(frag)
    return collapsed


def id_lookup(l):
    """Create dictionary where each unique item is key, value is the item numerical ID"""
    temp = defaultdict(lambda: len(temp))
    idx = [temp[x] for x in l]
    lookup = dict(zip(set(l), set(idx)))
    return lookup


def getFragments(
    interval,
    bam,
    min_mapq=30,
    cellbarcode="CB",
    readname_barcode=None,
    cells=None,
    max_distance=5000,
    min_distance=10,
    chunksize=500000,
    shifts=[4, -5],
    collapse_within=False
):
    """Extract ATAC fragments from BAM file

    Iterate over paired reads in a BAM file and extract the ATAC fragment coordinates

    Parameters
    ----------
    interval : list
        Genomic coordinates to iterate over. Must be a single region.
    bam : str
        Path to BAM file
    min_mapq : int
        Minimum MAPQ to retain fragment
    cellbarcode : str
        Tag used for cell barcode. Default is CB (used by cellranger)
    readname_barcode : str, optional
        Regex to extract cell barcode from readname. If None,
        use the read tag instead.
    cells : list, optional
        List of cell barocodes to retain
    max_distance : int, optional
        Maximum distance between integration sites for the fragment to be retained.
        Allows filtering of implausible fragments that likely result from incorrect
        mapping positions. Default is 5000 bp.
    min_distance : int, optional
        Minimum distance between integration sites for the fragment to be retained.
        Allows filtering of implausible fragments that likely result from incorrect
        mapping positions. Default is 10 bp.
    chunksize : int
        Number of BAM entries to read through before collapsing and writing
        fragments to disk. Higher chunksize will use more memory but will be
        faster.
    shifts : list
        Fragment position shifts to apply. First element defines shift for + strand,
        second element defines shift for - strand.
    collapse_within : bool
        Only collapse fragments containing the same start and end coordinate within the
        same cell barcode. Setting to True will only collapse duplicates if the cell barcode
        is the same (allows same fragment coordinates with different cell barcode)
    """
    fragment_dict = dict()
    inputBam = pysam.AlignmentFile(bam, "rb")
    outfile = tempfile.NamedTemporaryFile(delete=False)
    outname = outfile.name
    x = 0
    if readname_barcode is not None:
        readname_barcode = re.compile(readname_barcode)
    for i in inputBam.fetch(interval[0], 0, interval[1]):
        fragment_dict = updateFragmentDict(
            fragments=fragment_dict,
            segment=i,
            min_mapq=min_mapq,
            cellbarcode=cellbarcode,
            readname_barcode=readname_barcode,
            cells=cells,
            max_dist=max_distance,
            min_dist=min_distance,
            shifts=shifts
        )
        x += 1
        if x > chunksize:

            current_position = i.reference_start
            complete = findCompleteFragments(
                fragments=fragment_dict,
                max_dist=max_distance,
                current_position=current_position,
                max_collapse_dist=20,
            )
            collapsed = collapseFragments(fragments=complete, collapse_within=collapse_within)
            writeFragments(fragments=collapsed, filepath=outname)
            x = 0
            gc.collect()
    # end of chromosome
    # collapse and write the remaining fragments
    complete = findCompleteFragments(
        fragments=fragment_dict,
        max_dist=max_distance,
        current_position=math.inf, # ensure all fragments are collapsed
        max_collapse_dist=-max_distance,
    )
    collapsed = collapseFragments(fragments=complete, collapse_within=collapse_within)
    writeFragments(fragments=collapsed, filepath=outname)
    return outname


def findCompleteFragments(fragments, max_dist, current_position, max_collapse_dist=20):
    """Find complete fragments that are >max_dist bp away from
    the current BAM file position

    Parameters
    ----------
    fragments : dict
        A dictionary containing ATAC fragment information
    max_dist : int
        The maximum allowed distance between fragment start and
        end positions
    current_position : int
        The current position being looked at in the position-sorted
        BAM file
    max_collapse_dist : int
        Maximum allowed distance for fragments from the same cell
        barcode that share one Tn5 integration site (fragment
        start or end coordinate) to be collapsed into a single
        fragment.

    Moves completed fragments to a new dictionary
    Completed fragments will be deleted from the original dictionary
    """
    allkey = list(fragments.keys())
    completed = dict()
    d = max_dist + max_collapse_dist
    for key in allkey:
        if fragments[key][4]:  # complete fragment
            if (fragments[key][2] + d) < current_position:
                completed[key] = fragments[key][:-1]  # removes "completed" T/F information
                del fragments[key]
        else:
            # remove incomplete fragments that are
            # too far away to ever be complete
            if fragments[key][1] is None:
                if (fragments[key][2] + d) < current_position:
                    del fragments[key]
            elif fragments[key][2] is None:
                if (fragments[key][1] + d) < current_position:
                    del fragments[key]
            else:
                # start and end coordinates present without a cell barcode
                del fragments[key]
    return completed


def updateFragmentDict(
    fragments, segment, min_mapq, cellbarcode, readname_barcode, cells, max_dist, min_dist,
    shifts=[4, -5]
):
    """Update dictionary of ATAC fragments
    Takes a new aligned segment and adds information to the dictionary,
    returns a modified version of the dictionary

    Positions are 0-based
    Reads aligned to the + strand are shifted +4 bp (configurable by shifts)
    Reads aligned to the - strand are shifted -5 bp (configurable by shifts)

    Parameters
    ----------
    fragments : dict
        A dictionary containing ATAC fragment information
    segment : pysam.AlignedSegment
        An aligned segment
    min_mapq : int
        Minimum MAPQ to retain fragment
    cellbarcode : str
       Tag used for cell barcode. Default is CB (used by cellranger)
    readname_barcode : regex
        A compiled regex for matching cell barcode in read name. If None,
        use the read tags.
    cells : list
        List of cells to retain. If None, retain all cells found.
    max_dist : int
        Maximum allowed distance between fragment start and end sites
    min_dist : int
        Minimum allowed distance between fragment start and end sites
    shifts : list
        List of adjustments made to fragment position. First element defines + strand
        shift, second element defines - strand shift.
    """
    # because the cell barcode is not stored with each read pair (only one of the pair)
    # we need to look for each read separately rather than using the mate cigar / mate postion information
    if readname_barcode is not None:
        re_match = readname_barcode.search(segment.qname)
        cell_barcode = re_match.group()
    else:
        cell_barcode, _ = utils.scan_tags(segment.tags, cb=cellbarcode)
    if cells is not None and cell_barcode is not None:
        if cell_barcode not in cells:
            return fragments
    mapq = segment.mapping_quality
    if mapq < min_mapq:
        return fragments
    chromosome = segment.reference_name
    qname = segment.query_name
    rstart = segment.reference_start
    rend = segment.reference_end
    is_reverse = segment.is_reverse
    if (rend is None) or (rstart is None):
        return fragments
    # correct for 9 bp Tn5 shift
    if is_reverse:
        rend = rend + shifts[1]
    else:
        rstart = rstart + shifts[0]
    fragments = addToFragments(
        fragments, qname, chromosome, rstart, rend, cell_barcode, is_reverse, max_dist, min_dist
    )
    return fragments


def addToFragments(
    fragments, qname, chromosome, rstart, rend, cell_barcode, is_reverse, max_dist, min_dist
):
    """Add new fragment information to dictionary

    Parameters
    ----------

    fragments : dict
        A dictionary containing all the fragment information
    qname : str
        Read name
    chromosome : str
        Chromosome name
    rstart : int
        Alignment start position
    rend : int
        Alignment end position
    cell_barcode : str
        Cell barcode sequence
    is_reverse : bool
        Read is aligned to reverse strand
    max_dist : int
        Maximum allowed fragment size
    min_dist : int
        Minimum allowed fragment size
    """
    if qname in fragments.keys():
        if is_reverse:
            current_coord = fragments[qname][1]
            if current_coord is None:
                # read aligned to the wrong strand, don't include
                del fragments[qname]
            elif ((rend - current_coord) > max_dist) or ((rend - current_coord) < min_dist):
                # too far away, don't include
                del fragments[qname]
            else:
                if cell_barcode is None and fragments[qname][3] is None:
                    # both fragment ends present but no cell barcode
                    del fragments[qname]
                else:
                    if fragments[qname][3] is None:
                        fragments[qname][3] = cell_barcode
                    fragments[qname][2] = rend
                    fragments[qname][4] = True
        else:
            current_coord = fragments[qname][2]
            if current_coord is None:
                del fragments[qname]
            elif ((current_coord - rstart) > max_dist) or (
                (current_coord - rstart) < min_dist
            ):
                del fragments[qname]
            else:
                if cell_barcode is None and fragments[qname][3] is None:
                    del fragments[qname]
                else:
                    if fragments[qname][3] is None:
                        fragments[qname][3] = cell_barcode
                    fragments[qname][1] = rstart
                    fragments[qname][4] = True
    else:
        # new read pair, add to dictionary
        fragments[qname] = [
            chromosome,     # chromosome    0
            None,           # start         1
            None,           # end           2
            cell_barcode,   # cell          3
            False,          # complete      4
        ]
        if is_reverse:
            fragments[qname][2] = rend
        else:
            fragments[qname][1] = rstart
    return fragments


def fragments(
    bam,
    fragment_path,
    min_mapq=30,
    nproc=1,
    cellbarcode="CB",
    chromosomes="(?i)^chr",
    readname_barcode=None,
    cells=None,
    max_distance=5000,
    min_distance=10,
    chunksize=500000,
    shifts=[4, -5],
    collapse_within=False
):
    """Create ATAC fragment file from BAM file

    Iterate over reads in BAM file, extract fragment coordinates and cell barcodes.
    Collapse sequencing duplicates.

    Parameters
    ----------
    bam : str
        Path to BAM file
    fragment_path : str
        Path for output fragment file
    min_mapq : int
        Minimum MAPQ to retain fragment
    nproc : int, optional
        Number of processors to use. Default is 1.
    cellbarcode : str
       Tag used for cell barcode. Default is CB (used by cellranger)
    chromosomes : str, optional
        Regular expression used to match chromosome names to include in the
        output file. Default is "(?i)^chr" (starts with "chr", case-insensitive).
        If None, use all chromosomes in the BAM file.
    readname_barcode : str, optional
        Regular expression used to match cell barocde stored in read name.
        If None (default), use read tags instead. Use "[^:]*" to match all characters
        before the first colon (":").
    cells : str
        File containing list of cell barcodes to retain. If None (default), use all cell barcodes
        found in the BAM file.
    max_distance : int, optional
        Maximum distance between integration sites for the fragment to be retained.
        Allows filtering of implausible fragments that likely result from incorrect
        mapping positions. Default is 5000 bp.
    min_distance : int, optional
        Minimum distance between integration sites for the fragment to be retained.
        Allows filtering implausible fragments that likely result from incorrect
        mapping positions. Default is 10 bp.
    chunksize : int
        Number of BAM entries to read through before collapsing and writing
        fragments to disk. Higher chunksize will use more memory but will be
        faster.
    shifts : list
        Fragment position shifts to apply. First element defines shift for + strand,
        second element defines shift for - strand.
    collapse_within : bool
        Only collapse fragments containing the same start and end coordinate within the
        same cell barcode. Setting to True will only collapse duplicates if the cell barcode
        is the same (allows same fragment coordinates with different cell barcode)
    """
    nproc = int(nproc)
    chrom = utils.get_chromosomes(bam, keep_contigs=chromosomes)
    cells = utils.read_cells(cells)
    p = Pool(nproc)
    frag_lists = [
        p.map_async(
            functools.partial(
                getFragments,
                bam=bam,
                min_mapq=int(min_mapq),
                cellbarcode=cellbarcode,
                readname_barcode=readname_barcode,
                cells=cells,
                max_distance=max_distance,
                min_distance=min_distance,
                chunksize=chunksize,
                shifts=shifts,
                collapse_within=collapse_within
            ),
            list(chrom.items()),
        )
    ]
    filenames = [res.get() for res in frag_lists]
    # cat files and write to output
    with open(fragment_path, "w") as outfile:
        for i in filenames:
            for j in i:
                with open(j, "r") as infile:
                    for line in infile:
                        outfile.write(line)
                os.remove(j)


import gzip
import os


def a2b(s, inverse=False):
    # just a silly function to convert ascii <-> binary
    if inverse:
        return s.decode('ascii')
    return bytes(s, encoding='ascii')


def correct_barcodes(barcodes, whitelist):
    """
    Use a whitelist computed with some external tool (such as UMI-tools)
    to correct the barcodes. This could be useful whenever the fastq file with 
    cell barcodes does not contain corrected barcodes.
    """
    if type(whitelist) == dict:
        # whitelist is derived from UMI-tools
        # create a dict where everything is corrected to itself
        u_barcodes = set(barcodes)
        corrected = dict(zip(u_barcodes, u_barcodes))
        # iterate over whitelist to add actual corrections where possible
        for bc in whitelist:
            for rbc in whitelist[bc]:
                corrected[rbc] = bc
    else:
        # whitelist is just a list of cell barcodes, hence
        # we need to perform corrections
        from umi_tools import UMIClusterer
    
        counts = dict()
    
        for bc in set(barcodes):
            # add every barcode
            # convert to bytes as UMIClusterer expects
            counts[a2b(bc)] = 1
    
        # add whitelist to the counter, making it the most abundant so that
        # UMIClusterer makes them the "top"
        for bc in whitelist:
            if not 'N' in bc:
                counts[a2b(bc)] = 1000 # I guess "2" works as well :-)
                
        clusterer = UMIClusterer(cluster_method='directional')
        
        corrected = dict.fromkeys(barcodes)
        
        for entry in clusterer(counts, threshold=1):
            for bc in entry:
                # assign every sequence the first in the list
                corrected[a2b(bc, inverse=True)] = a2b(entry[0], inverse=True)

    # return the list of corrected
    return [corrected[bc] for bc in barcodes]
    
    
def sniff_whitelist(filename, nr=100):
    # possibly the dumbest function ever written
    nl = 0 # number of lines sniffed
    nc = 0 # max number of corrected bc found
    nf = 0 # max number of fields
    with open(filename) as fh:
        while nl < nr:
            line = fh.readline()
            nl += 1
            l = len(line.split())
            if l > nf:
                nf = l
            if l > 1:
                l = len(line.split()[1].split(','))
                if l > nc:
                    nc = l
    if nf > 2 and nc > 1:
        # assuming a UMI-tools whitelist contains at least
        # three columns and the second one contains at least
        # two corrected barcodes
        return 1
    return 0

def addbarcodes(cb_position, fq1, fq2, fq3=None, prefix="", suffix="", wl=None):
    """Add cell barcode to read names

    Parameters
    ----------
    cb_position : int
        Base positions containing cell barcode information
    fq1 : str
        Fastq file containing cell barcode sequences
    fq2 : str
        Fastq file to add cell barcodes to
    fq3 : str
        Fastq file to add cell barcodes to (optional)
    prefix : str
        Prefix to append to cell barcodes
    suffix : str
        Suffix to append to cell barcodes
    """
    barcodes, whitelist = get_barcodes(f=fq1, bases=cb_position, prefix=prefix, suffix=suffix, wl=wl)
    if len(whitelist) > 0:
        barcodes = correct_barcodes(barcodes, whitelist)
    add_barcodes(f=fq2, cb=barcodes)
    if fq3 is not None:
        add_barcodes(f=fq3, cb=barcodes)


def get_barcodes(f, bases=12, prefix="", suffix="", wl=None):
    f_open = open_fastq(f)
    if f.endswith(".gz"):
        gz = True
    else:
        gz = False
    x = 0
    cb = []
    for i in f_open:
        if (x % 4 == 1):
            if gz:
                cb.append(prefix + i.decode("utf-8")[:bases] + suffix)
            else:
                cb.append(prefix + i[:bases] + suffix)
        x += 1
    f_open.close()
    if wl is not None:
        # check if the whitelist already contains corrected barcodes
        wlt = sniff_whitelist(wl)
        if wlt == 1:
            whitelist = {}
        else:
            whitelist = []
        for line in open(wl):
            fields = line.split()
            if wlt == 1:
                whitelist[fields[0]] = fields[1].split(',')
            else:
                whitelist.append(fields[0])
    else:
        whitelist = None
    return(cb, whitelist)


def add_barcodes(f, cb):
    f_open = open_fastq(f)
    if f.endswith(".gz"):
        gz = True
        o = f.replace(".fastq.gz", "").replace(".fq.gz", "") + ".barcoded.fastq.gz"
        outfile = gzip.GzipFile(o, mode = "wb")
    else:
        gz = False
        o = f.replace(".fastq", "").replace(".fq", "") + ".barcoded.fastq"
        outfile = open(o, "w+")
    x = 0
    y = 0
    for i in f_open:
        if (x % 4 ==0):
            if gz:
                rdname = i.decode("utf-8")
            else:
                rdname = i
            rdname = "@" + cb[y] + ":" + rdname[1:]
            if gz:
                outfile.write(rdname.encode("utf-8"))
            else:
                outfile.write(rdname)
            y += 1
        else:
            outfile.write(i)
        x += 1
    f_open.close()
    outfile.close()


def open_fastq(f):
    if os.path.isfile(f):
        if f.endswith(".gz"):
            f_open = gzip.open(f, "rb")
        else:
            f_open = open(f, "r")
        return(f_open)
    else:
        raise Exception("File not found")



import pytest
import os
from sinto import addtags


@pytest.mark.parametrize("nproc", [1, 2])
def test_addtags(tmpdir, nproc):
    basepath=os.path.dirname(os.path.realpath(__file__))
    bam=os.path.join(basepath, "data/test.bam")
    tags=os.path.join(basepath, "data/tags.tsv")
    outf=os.path.join(tmpdir, "tags.sam")
    addtags.addtags(
        bam=bam,
        tagfile=tags,
        trim_suffix=False,
        output=outf,
        sam=True,
        nproc=nproc,
        mode="tag"
    )
    
    assert os.path.exists(outf)

    with open(outf, 'r') as inp:
        for l in inp:
            if l.startswith("@"):
                next
            else:
                break
    with open(os.path.join(basepath, "data/tags_line1.sam"), 'r') as inp:
        for l2 in inp:
            next
    assert l == l2


import pytest
import os
from sinto import filterbarcodes


@pytest.mark.parametrize("nproc", [1, 2])
def test_filterbarcodes(tmpdir, nproc):
    basepath=os.path.dirname(os.path.realpath(__file__))
    bam=os.path.join(basepath, "data/test.bam")
    barcodes=os.path.join(basepath, "data/barcodes.tsv")

    filterbarcodes.filterbarcodes(
        cells=barcodes,
        bam=bam,
        trim_suffix=False,
        nproc=nproc,
        readname_barcode=None,
        cellbarcode="CB",
        outdir=tmpdir,
        sam=True
    )

    exists = [os.path.exists(os.path.join(tmpdir, x)) for x in ['A.sam', 'B.sam', 'C.sam']]
    assert all(exists)
    
    # check contents
    with open(os.path.join(tmpdir, 'A.sam'), 'r') as inp:
        for l in inp:
            if l.startswith("@"):
                next
            else:
                break
    with open(os.path.join(basepath, "data/a_line1.sam"), 'r') as inp:
        for l2 in inp:
            next

    assert l == l2


import subprocess


def test_adds_CB_tag_to_rg(tmpdir):

    # given
    # modified from an example in the SAM specification v1
    sam = (
        "@HD VN:1.5 SO:coordinate\n"
        "@SQ SN:20 LN:63025520\n"
        "@RG ID:rg1 SM:sample_1 LB:1 PU:1 PL:ILLUMINA\n"
        "r002 0 20 9 30 3S6M1P1I4M * 0 0 AAAAGATAAGGATA * CB:Z:AAAA-1 RG:Z:rg1\n"
    ).replace(" ", "\t")
    in_sam_file = tmpdir / "in.sam"
    in_tag_file = tmpdir / "tags.txt"
    output = tmpdir/'output.sam'
    with open(in_sam_file, "w") as fh:
        fh.write(sam)
    with open(in_tag_file, "w") as fh:
        fh.write("AAAA-1")

    # when
    subprocess.run(
        f"sinto tagtorg --tagfile {in_tag_file} --tag CB -b - -o - < {in_sam_file} > {output}",
        shell=True,
    )

    # then
    out_sam = open(output).read()
    expected = (
        "@HD VN:1.5 SO:coordinate\n"
        "@SQ SN:20 LN:63025520\n"
        "@RG ID:rg1 SM:sample_1 LB:1 PU:1 PL:ILLUMINA\n"
        "@RG ID:rg1:AAAA-1 SM:sample_1:AAAA-1 LB:1 PU:1 PL:ILLUMINA\n"
        "r002 0 20 9 30 3S6M1P1I4M * 0 0 AAAAGATAAGGATA * CB:Z:AAAA-1 RG:Z:rg1:AAAA-1\n"
    ).replace(" ", "\t")
    assert expected == out_sam

def test_ignores_unknown_cb_tag(tmpdir):

    # given
    # modified from an example in the SAM specification v1
    sam = (
        "@HD VN:1.5 SO:coordinate\n"
        "@SQ SN:20 LN:63025520\n"
        "@RG ID:rg1 SM:sample_1 LB:1 PU:1 PL:ILLUMINA\n"
        "r002 0 20 9 30 3S6M1P1I4M * 0 0 AAAAGATAAGGATA * CB:Z:AAAA-1 RG:Z:rg1\n"
        "r002 0 20 9 30 3S6M1P1I4M * 0 0 AAAAGATAAGGATA * CB:Z:CCCC-1 RG:Z:rg1\n"
    ).replace(" ", "\t")
    in_sam_file = tmpdir / "in.sam"
    in_tag_file = tmpdir / "tags.txt"
    output = tmpdir/'output.sam'
    with open(in_sam_file, "w") as fh:
        fh.write(sam)
    with open(in_tag_file, "w") as fh:
        fh.write("AAAA-1")

    # when
    subprocess.run(
        f"sinto tagtorg --tagfile {in_tag_file} --tag CB -b - -o - < {in_sam_file} > {output}",
        shell=True,
    )

    # then
    out_sam = open(output).read()
    expected = (
        "@HD VN:1.5 SO:coordinate\n"
        "@SQ SN:20 LN:63025520\n"
        "@RG ID:rg1 SM:sample_1 LB:1 PU:1 PL:ILLUMINA\n"
        "@RG ID:rg1:AAAA-1 SM:sample_1:AAAA-1 LB:1 PU:1 PL:ILLUMINA\n"
        "r002 0 20 9 30 3S6M1P1I4M * 0 0 AAAAGATAAGGATA * CB:Z:AAAA-1 RG:Z:rg1:AAAA-1\n"
        "r002 0 20 9 30 3S6M1P1I4M * 0 0 AAAAGATAAGGATA * CB:Z:CCCC-1 RG:Z:rg1\n"
    ).replace(" ", "\t")
    assert expected == out_sam



from sinto import (
    utils,
    filterbarcodes,
    addtags,
    fragments,
    tagtorg,
    addbarcodes,
    tagtotag,
    tagtoname,
    blocks
)


@utils.log_info
def run_filterbarcodes(options):
    """Wraps the sinto.filterbarcodes function for use on the command line
    """
    filterbarcodes.filterbarcodes(
        cells=options.cells,
        bam=options.bam,
        trim_suffix=options.trim_suffix,
        nproc=options.nproc,
        readname_barcode=options.barcode_regex,
        cellbarcode=options.barcodetag,
        outdir=options.outdir,
        sam=options.sam
    )


@utils.log_info
def run_addtags(options):
    """Wraps the sinto.addtags function for use on the command line
    """
    addtags.addtags(
        bam=options.bam,
        tagfile=options.tagfile,
        trim_suffix=options.trim_suffix,
        output=options.output,
        sam=options.sam,
        nproc=options.nproc,
        mode=options.mode,
    )


@utils.log_info
def run_fragments(options):
    """Wraps the sinto.fragments function for use on the command line
    """
    fragments.fragments(
        bam=options.bam,
        fragment_path=options.fragments,
        min_mapq=options.min_mapq,
        nproc=options.nproc,
        cellbarcode=options.barcodetag,
        readname_barcode=options.barcode_regex,
        chromosomes=options.use_chrom,
        cells=options.cells,
        max_distance=options.max_distance,
        min_distance=options.min_distance,
        chunksize=options.chunksize,
        shifts=[options.shift_plus, options.shift_minus],
        collapse_within=options.collapse_within
    )


@utils.log_info
def run_tagtorg(options):
    tagtorg.tagtorg(
        bam=options.bam,
        tag_value_file=options.tagfile,
        tag=options.tag,
        output=options.output,
        out_format=options.outputformat,
    )


@utils.log_info
def run_blocks(options):
    blocks.blocks(
        bam=options.bam,
        block_path=options.blocks,
        min_mapq=options.min_mapq,
        nproc=options.nproc,
        cellbarcode=options.barcodetag,
        umibarcode=options.umitag,
        readname_barcode=options.barcode_regex,
        cells=options.cells,
        include_strand=options.strand,
    )

@utils.log_info
def run_tagtotag(options):
    tagtotag.tagtotag(
        bam=options.bam,
        from_tag=options.from_,
        to_tag=options.to,
        output=options.output,
        delete=options.delete,
        out_format=options.outputformat,
    )


@utils.log_info
def run_barcode(options):
    addbarcodes.addbarcodes(
        cb_position=options.bases,
        fq1=options.barcode_fastq,
        fq2=options.read1,
        fq3=options.read2,
        prefix=options.prefix,
        suffix=options.suffix,
        wl=options.whitelist,
    )

@utils.log_info
def run_tagtoname(options):
    tagtoname.move(
        bam=options.bam,
        output=options.output,
        cb_tag=options.tag,
        out_format=options.outputformat,
        from_tag=True
    )

@utils.log_info
def run_nametotag(options):
    tagtoname.move(
        bam=options.bam,
        output=options.output,
        cb_tag=options.tag,
        out_format=options.outputformat,
        from_tag=False,
        cb_regex=options.barcode_regex
    )



import pysam
from sinto import utils
import re
import os
from multiprocessing import Pool
import functools
import tempfile


def getBlocks(
    intervals,
    bam,
    min_mapq=30,
    cellbarcode="CB",
    umibarcode="UB",
    readname_barcode=None,
    cells=None,
    include_strand=False
):
    inputBam = pysam.AlignmentFile(bam, "rb")
    outfile = tempfile.NamedTemporaryFile(delete=False)
    outname = outfile.name
    outf = open(outname, "w+")
    for i in intervals:
        for r in inputBam.fetch(i[0], i[1], i[2]):
            if r.mapping_quality < min_mapq:
                continue
            if readname_barcode is not None:
                re_match = readname_barcode.match(r.qname)
                cell_barcode = re_match.group()
                _, umi = utils.scan_tags(r.tags, ub=umibarcode)
            else:
                cell_barcode, umi = utils.scan_tags(r.tags, cb=cellbarcode, ub=umibarcode)
            if cell_barcode is None or umi is None:
                continue
            if (cells is not None) and (cell_barcode not in cells):
                continue
            blocks = r.get_blocks()
            chrom = r.reference_name
            strand = "-" if r.is_reverse else "+"
            for i in blocks:
                outstr = "\t".join(map(str, [chrom, i[0], i[1], cell_barcode, umi]))
                if include_strand:
                    outstr += "\t" + strand
                outf.write(outstr + "\n")
    outf.close()
    inputBam.close()
    return(outname)


def blocks(
    bam,
    block_path,
    min_mapq=30,
    nproc=1,
    cellbarcode="CB",
    umibarcode="UB",
    readname_barcode=None,
    cells=None,
    include_strand=False,
):
    """Create scRNA-seq block file from BAM file

    Iterate over all aligned reads in a BAM file and extract the 
    coordinates of each aligned block. Each aligned block is written
    as a separate entry in a BED file, along with the cell barcode
    and UMI sequence for the read.

    Parameters
    ----------
    bam : str
        Path to BAM file
    block_path : str
        Path for output block file
    min_mapq : int
        Minimum MAPQ to retain fragment
    nproc : int, optional
        Number of processors to use. Default is 1.
    cellbarcode : str
        Tag used for cell barcode. Default is CB (used by cellranger)
    umibarcode : str
        Tag used for the UMI sequence. Defauly is UB (used by cellranger)
    readname_barcode : str, optional
        Regular expression used to match cell barocde stored in read name. 
        If None (default), use read tags instead. Use "[^:]*" to match all characters 
        before the first colon (":").
    cells : str
        File containing list of cell barcodes to retain. If None (default), use all cell barcodes
        found in the BAM file.
    include_strand : bool
        Include strand information
    """
    nproc = int(nproc)
    inputBam = pysam.AlignmentFile(bam, "rb")
    intervals = utils.chunk_bam(inputBam, nproc, unmapped=False)
    inputBam.close()
    if readname_barcode is not None:
        readname_barcode = re.compile(readname_barcode)
    cells = utils.read_cells(cells)
    p = Pool(nproc)
    tmpfiles = [
        p.map_async(
            functools.partial(
                getBlocks,
                bam=bam,
                min_mapq=int(min_mapq),
                cellbarcode=cellbarcode,
                umibarcode=umibarcode,
                readname_barcode=readname_barcode,
                cells=cells,
                include_strand=include_strand,
            ),
            intervals.values(),
        )
    ]
    filenames = [res.get() for res in tmpfiles]
    # cat files and write to output
    with open(block_path, "w") as outfile:
        for i in filenames:
            for j in i:
                with open(j, "r") as infile:
                    for line in infile:
                        outfile.write(line)
                os.remove(j)


import subprocess

import pytest


@pytest.mark.parametrize("delete", [True, False])
def test_copies_CB_tag_to_cb(tmpdir, delete):

    # given
    # modified from an example in the SAM specification v1
    sam = (
        "@HD VN:1.5 SO:coordinate\n"
        "@SQ SN:20 LN:63025520\n"
        "r002 0 20 9 30 3S6M1P1I4M * 0 0 AAAAGATAAGGATA * CB:Z:AAAA-1\n"
    ).replace(" ", "\t")
    in_sam_file = tmpdir / "in.sam"
    in_tag_file = tmpdir / "tags.txt"
    output = tmpdir / "output.sam"
    with open(in_sam_file, "w") as fh:
        fh.write(sam)
    with open(in_tag_file, "w") as fh:
        fh.write("AAAA-1")

    # when
    cmd = f"sinto tagtotag --from CB --to xx --bam - -o -"
    if delete:
        cmd += " --delete"
    cmd += f" < {in_sam_file} > {output}"
    subprocess.run(cmd, shell=True)

    # then
    out_sam = open(output).read()
    if delete:
        expected = (
            "@HD VN:1.5 SO:coordinate\n"
            "@SQ SN:20 LN:63025520\n"
            "r002 0 20 9 30 3S6M1P1I4M * 0 0 AAAAGATAAGGATA * xx:Z:AAAA-1\n"
        )

    else:
        expected = (
            "@HD VN:1.5 SO:coordinate\n"
            "@SQ SN:20 LN:63025520\n"
            "r002 0 20 9 30 3S6M1P1I4M * 0 0 AAAAGATAAGGATA * CB:Z:AAAA-1 xx:Z:AAAA-1\n"
        )
    expected = expected.replace(" ", "\t")
    assert expected == out_sam

def test_copies_CB_tag_to_cb_from_bam(tmpdir):

    # given
    # modified from an example in the SAM specification v1
    sam = (
        "@HD VN:1.5 SO:coordinate\n"
        "@SQ SN:20 LN:63025520\n"
        "r002 0 20 9 30 3S6M1P1I4M * 0 0 AAAAGATAAGGATA * CB:Z:AAAA-1\n"
    ).replace(" ", "\t")
    in_sam_file = tmpdir / "in.sam"
    in_tag_file = tmpdir / "tags.txt"
    output = tmpdir / "output.sam"
    with open(in_sam_file, "w") as fh:
        fh.write(sam)
    with open(in_tag_file, "w") as fh:
        fh.write("AAAA-1")
    subprocess.run(f"samtools view {in_sam_file} -OBAM -o {in_sam_file}.binary-sam --no-PG", shell=True)

    # when
    cmd = f"sinto tagtotag --from CB --to xx --bam {in_sam_file}.binary-sam -o {output}"
    subprocess.run(cmd, shell=True)

    # then
    out_sam = open(output).read()
    expected = (
        "@HD VN:1.5 SO:coordinate\n"
        "@SQ SN:20 LN:63025520\n"
        "r002 0 20 9 30 3S6M1P1I4M * 0 0 AAAAGATAAGGATA * CB:Z:AAAA-1 xx:Z:AAAA-1\n"
    )
    expected = expected.replace(" ", "\t")
    assert expected == out_sam




import pysam
from multiprocessing import Pool
import functools
import random
import string
from sinto import utils
import re
import os
from itertools import chain
from pathlib import Path


def _iterate_reads(
    intervals,
    bam,
    cb,
    classes,
    trim_suffix,
    cellbarcode,
    readname_barcode,
    sam,
    outdir
):
    inputBam = pysam.AlignmentFile(bam, "rb")
    header = inputBam.header.to_dict()
    validKeys = [x for x in ["HD", "SQ", "RG"] if x in header.keys()]
    newhead = dict((k, header[k]) for k in validKeys)
    ident = "".join(
        random.choice(string.ascii_uppercase + string.digits) for _ in range(6)
    )
    filemode = 'w' if sam else 'wb'
    filelist = [os.path.join(outdir, x + "_" + ident) for x in classes]
    bamlist = [pysam.AlignmentFile(x, filemode, header=newhead) for x in filelist]
    for i in intervals:
        for r in inputBam.fetch(i[0], i[1], i[2]):
            if readname_barcode is not None:
                re_match = readname_barcode.search(r.qname)
                cell_barcode = re_match.group()
            else:
                cell_barcode, _ = utils.scan_tags(r.tags, cb=cellbarcode)
            if cell_barcode is not None:
                if trim_suffix:
                    cell_barcode = cell_barcode[:-2]
                if cell_barcode in cb.keys():
                    cell_classes = cb[cell_barcode]
                    for j in cell_classes:
                        fileindex = filelist.index(os.path.join(outdir, j + "_" + ident))
                        bamlist[fileindex].write(r)
    for i in bamlist:
        i.close()
    inputBam.close()
    return ident


def mergeAll(original, idents, classes, nproc, outdir, sam, remove=True):
    """Merge all temp files for each class
    If remove is True, remove temp files after
    successful merging"""
    suffix = '.sam' if sam else '.bam'
    for i in classes:
        allfiles = [os.path.join(outdir, i + "_" + x) for x in idents]
        output = os.path.join(outdir, i + suffix)
        pysam.merge('-@', str(nproc), '--no-PG', '-h', original, '-c', output, *allfiles)
        if remove:
            if os.path.exists(output):
                [os.remove(i) for i in allfiles]
            else:
                raise Exception("samtools merge failed, temp files not deleted")


def filterbarcodes(
    cells, bam, readname_barcode, cellbarcode, outdir, sam=False, trim_suffix=True, nproc=1
):
    """Filter reads based on input list of cell barcodes

    Copy BAM entries matching a list of cell barcodes to a new BAM file.
    Output BAM files will be named according to the group name in the 
    file provided.

    Parameters
    ----------
    cells : str
        Path to file containing cell barcodes and the group associated with each barcode.
        File can be gzip compressed. A separate BAM file will be created for each 
        group of cells.
    bam : str
        Path to BAM file.
    trim_suffix: bool, optional
        Remove trailing 2 characters from cell barcode in bam file (sometimes needed to match 10x barcodes).
    nproc : int, optional
        Number of processors to use. Default is 1.
    cellbarcode : str
       Tag used for cell barcode. Default is CB (used by cellranger)
    readname_barcode : regex
        A regular expression for matching cell barcode in read name. If None (default),
        use the read tags.

    Raises
    ------
    Exception
        If samtools merge of temporary BAM files fails
    """
    nproc = int(nproc)
    cb = utils.read_cell_barcode_file(cells)
    unique_classes = list(set(chain.from_iterable(cb.values())))
    inputBam = pysam.AlignmentFile(bam, "rb")
    intervals = utils.chunk_bam(inputBam, nproc, unmapped=True)
    ident = "".join(random.choice(string.ascii_uppercase + string.digits) for _ in range(6))
    inputBam.close()
    Path(outdir).mkdir(parents=True, exist_ok=True)
    if readname_barcode is not None:
        readname_barcode = re.compile(readname_barcode)
    p = Pool(nproc)
    idents = p.map_async(
        functools.partial(
            _iterate_reads,
            bam=bam,
            cb=cb,
            classes=unique_classes,
            trim_suffix=trim_suffix,
            cellbarcode=cellbarcode,
            readname_barcode=readname_barcode,
            outdir=outdir,
            sam=sam
        ),
        intervals.values(),
    ).get(9999999)
    mergeAll(
        original=bam,
        idents=idents,
        classes=unique_classes,
        nproc=nproc,
        outdir=outdir,
        sam=sam,
        remove=True
    )



import pytest
import os
from sinto import fragments


@pytest.mark.parametrize("collapse_within", [True, False])
@pytest.mark.parametrize("nproc", [1, 2])
def test_fragments(tmpdir, collapse_within, nproc):
    basepath=os.path.dirname(os.path.realpath(__file__))
    bam=os.path.join(basepath, "data/test.bam")
    outf=tmpdir / "frags.bed"
    fragments.fragments(
        bam=bam,
        fragment_path=outf,
        collapse_within=collapse_within,
        nproc=nproc
    )
    output = [line.strip("\n") for line in open(outf, "r")]    
    if collapse_within:
        expected_path = os.path.join(basepath, "data/frags_within.bed")
    else:
        expected_path = os.path.join(basepath, "data/frags_across.bed")
    expected = [line.strip("\n") for line in open(expected_path, "r")]
    output.sort()
    expected.sort()
    assert output == expected
